# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.exceptions import UnsupportedTLSVersionWarning
from tests import BaseSessionTest, mock


class TestOpensslVersion(BaseSessionTest):
    def test_incompatible_openssl_version(self):
        with mock.patch('ssl.OPENSSL_VERSION_INFO', new=(0, 9, 8, 11, 15)):
            with mock.patch('warnings.warn') as mock_warn:
                self.session.create_client('iot-data', 'us-east-1')
                call_args = mock_warn.call_args[0]
                warning_message = call_args[0]
                warning_type = call_args[1]
                # We should say something specific about the service.
                self.assertIn('iot-data', warning_message)
                self.assertEqual(warning_type, UnsupportedTLSVersionWarning)

    def test_compatible_openssl_version(self):
        with mock.patch('ssl.OPENSSL_VERSION_INFO', new=(1, 0, 1, 1, 1)):
            with mock.patch('warnings.warn') as mock_warn:
                self.session.create_client('iot-data', 'us-east-1')
                self.assertFalse(mock_warn.called)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore.docs.params import (
    RequestParamsDocumenter,
    ResponseParamsDocumenter,
)
from botocore.docs.utils import DocumentedShape
from botocore.hooks import HierarchicalEmitter
from tests import mock
from tests.unit.docs import BaseDocsTest


class BaseParamsDocumenterTest(BaseDocsTest):
    def setUp(self):
        super().setUp()
        self.event_emitter = HierarchicalEmitter()
        self.request_params = RequestParamsDocumenter(
            service_name='myservice',
            operation_name='SampleOperation',
            event_emitter=self.event_emitter,
        )
        self.response_params = ResponseParamsDocumenter(
            service_name='myservice',
            operation_name='SampleOperation',
            event_emitter=self.event_emitter,
        )


class TestDocumentDefaultValue(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape_to_params('Foo', 'String', 'This describes foo.')

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [':type Foo: string', ':param Foo: This describes foo.']
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            ['- *(dict) --*', '  - **Foo** *(string) --* This describes foo.']
        )


class TestTraverseAndDocumentShape(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape_to_params('Foo', 'String', 'This describes foo.')
        self.event_emitter = mock.Mock()
        self.request_params = RequestParamsDocumenter(
            service_name='myservice',
            operation_name='SampleOperation',
            event_emitter=self.event_emitter,
        )
        self.response_params = ResponseParamsDocumenter(
            service_name='myservice',
            operation_name='SampleOperation',
            event_emitter=self.event_emitter,
        )

    def test_events_emitted_response_params(self):
        self.response_params.traverse_and_document_shape(
            section=self.doc_structure,
            shape=self.operation_model.input_shape,
            history=[],
        )
        self.assertEqual(
            self.event_emitter.emit.call_args_list,
            [
                mock.call(
                    'docs.response-params.myservice.SampleOperation.Foo',
                    section=self.doc_structure.get_section('Foo'),
                ),
                mock.call(
                    (
                        'docs.response-params.myservice.SampleOperation'
                        '.complete-section'
                    ),
                    section=self.doc_structure,
                ),
            ],
        )

    def test_events_emitted_request_params(self):
        self.request_params.traverse_and_document_shape(
            section=self.doc_structure,
            shape=self.operation_model.input_shape,
            history=[],
        )
        self.assertEqual(
            self.event_emitter.emit.call_args_list,
            [
                mock.call(
                    'docs.request-params.myservice.SampleOperation.Foo',
                    section=self.doc_structure.get_section('Foo'),
                ),
                mock.call(
                    (
                        'docs.request-params.myservice.SampleOperation'
                        '.complete-section'
                    ),
                    section=self.doc_structure,
                ),
            ],
        )


class TestDocumentMultipleDefaultValues(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape_to_params('Foo', 'String', 'This describes foo.')
        self.add_shape_to_params(
            'Bar', 'String', 'This describes bar.', is_required=True
        )

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                ':type Foo: string',
                ':param Foo: This describes foo.',
                ':type Bar: string',
                ':param Bar: **[REQUIRED]** This describes bar.',
            ]
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                '- *(dict) --*',
                '  - **Foo** *(string) --* This describes foo.',
                '  - **Bar** *(string) --* This describes bar.',
            ]
        )


class TestDocumentInclude(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape_to_params('Foo', 'String', 'This describes foo.')
        self.include_params = [
            DocumentedShape(
                name='Baz',
                type_name='integer',
                documentation='This describes baz.',
            )
        ]

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure,
            self.operation_model.input_shape,
            include=self.include_params,
        )
        self.assert_contains_lines_in_order(
            [
                ':type Foo: string',
                ':param Foo: This describes foo.',
                ':type Baz: int',
                ':param Baz: This describes baz.',
            ]
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure,
            self.operation_model.input_shape,
            include=self.include_params,
        )
        self.assert_contains_lines_in_order(
            [
                '- *(dict) --*',
                '  - **Foo** *(string) --* This describes foo.',
                '  - **Baz** *(integer) --* This describes baz.',
            ]
        )


class TestDocumentExclude(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape_to_params('Foo', 'String', 'This describes foo.')
        self.add_shape_to_params(
            'Bar', 'String', 'This describes bar.', is_required=True
        )
        self.exclude_params = ['Foo']

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure,
            self.operation_model.input_shape,
            exclude=self.exclude_params,
        )
        self.assert_contains_lines_in_order(
            [
                ':type Bar: string',
                ':param Bar: **[REQUIRED]** This describes bar.',
            ]
        )
        self.assert_not_contains_lines(
            [':type Foo: string', ':param Foo: This describes foo.']
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure,
            self.operation_model.input_shape,
            exclude=self.exclude_params,
        )
        self.assert_contains_lines_in_order(
            ['- *(dict) --*', '  - **Bar** *(string) --* This describes bar.']
        )
        self.assert_not_contains_line(
            '  - **Foo** *(string) --* This describes foo.'
        )


class TestDocumentList(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape(
            {
                'List': {
                    'type': 'list',
                    'member': {
                        'shape': 'String',
                        'documentation': 'A string element',
                    },
                }
            }
        )
        self.add_shape_to_params(
            'Foo',
            'List',
            'This describes the list. Each element of this list is a string.',
        )

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                ':type Foo: list',
                ':param Foo: This describes the list.',
                '  - *(string) --* A string element',
            ]
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                '- *(dict) --*',
                (
                    '  - **Foo** *(list) --* This describes the list. '
                    'Each element of this list is a string.'
                ),
                '    - *(string) --* A string element',
            ]
        )


class TestDocumentMap(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape(
            {
                'Map': {
                    'type': 'map',
                    'key': {'shape': 'String'},
                    'value': {'shape': 'String'},
                }
            }
        )
        self.add_shape_to_params('Foo', 'Map', 'This describes the map.')

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                ':type Foo: dict',
                ':param Foo: This describes the map.',
                '  - *(string) --*',
                '    - *(string) --*',
            ]
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                '- *(dict) --*',
                '  - **Foo** *(dict) --* This describes the map.',
                '    - *(string) --*',
                '      - *(string) --*',
            ]
        )


class TestDocumentStructure(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape(
            {
                'Structure': {
                    'type': 'structure',
                    'members': {
                        'Member': {
                            'shape': 'String',
                            'documentation': 'This is its member.',
                        }
                    },
                }
            }
        )
        self.add_shape_to_params(
            'Foo', 'Structure', 'This describes the structure.'
        )

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                ':type Foo: dict',
                ':param Foo: This describes the structure.',
                '  - **Member** *(string) --* This is its member.',
            ]
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                '- *(dict) --*',
                '  - **Foo** *(dict) --* This describes the structure.',
                '    - **Member** *(string) --* This is its member.',
            ]
        )


class TestDocumentRecursiveShape(BaseParamsDocumenterTest):
    def setUp(self):
        super().setUp()
        self.add_shape(
            {
                'Structure': {
                    'type': 'structure',
                    'members': {
                        'Foo': {
                            'shape': 'Structure',
                            'documentation': 'This is a recursive structure.',
                        }
                    },
                }
            }
        )
        self.add_shape_to_params(
            'Foo', 'Structure', 'This describes the structure.'
        )

    def test_request_params(self):
        self.request_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                ':type Foo: dict',
                ':param Foo: This describes the structure.',
                '  - **Foo** *(dict) --* This is a recursive structure.',
            ]
        )

    def test_response_params(self):
        self.response_params.document_params(
            self.doc_structure, self.operation_model.input_shape
        )
        self.assert_contains_lines_in_order(
            [
                '- *(dict) --*',
                '  - **Foo** *(dict) --* This is a recursive structure.',
            ]
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under thimport mock
import contextlib
import copy
import json
import logging
import os
import socket
import threading

import pytest

import botocore.config
import botocore.exceptions
import botocore.session
from botocore import xform_name
from tests import ClientHTTPStubber, mock, temporary_file

logger = logging.getLogger(__name__)

CASES_FILE = os.path.join(os.path.dirname(__file__), 'cases.json')
DATA_DIR = os.path.join(os.path.dirname(__file__), 'data/')


class RetryableException(botocore.exceptions.EndpointConnectionError):
    fmt = '{message}'


class NonRetryableException(Exception):
    pass


EXPECTED_EXCEPTIONS_THROWN = (
    botocore.exceptions.ClientError,
    NonRetryableException,
    RetryableException,
)


def _load_test_cases():
    with open(CASES_FILE) as f:
        loaded_tests = json.loads(f.read())
    test_cases = _get_cases_with_defaults(loaded_tests)
    _replace_expected_anys(test_cases)
    return test_cases


def _get_cases_with_defaults(loaded_tests):
    cases = []
    defaults = loaded_tests['defaults']
    for case in loaded_tests['cases']:
        base = copy.deepcopy(defaults)
        base.update(case)
        cases.append(base)
    return cases


def _replace_expected_anys(test_cases):
    for case in test_cases:
        for expected_event in case['expectedMonitoringEvents']:
            for entry, value in expected_event.items():
                if value in ['ANY_STR', 'ANY_INT']:
                    expected_event[entry] = mock.ANY


@pytest.mark.parametrize("test_case", _load_test_cases())
def test_client_monitoring(test_case):
    _run_test_case(test_case)


@contextlib.contextmanager
def _configured_session(case_configuration, listener_port):
    environ = {
        'AWS_ACCESS_KEY_ID': case_configuration['accessKey'],
        'AWS_SECRET_ACCESS_KEY': 'secret-key',
        'AWS_DEFAULT_REGION': case_configuration['region'],
        'AWS_DATA_PATH': DATA_DIR,
        'AWS_CSM_PORT': listener_port,
    }
    if 'sessionToken' in case_configuration:
        environ['AWS_SESSION_TOKEN'] = case_configuration['sessionToken']
    environ.update(case_configuration['environmentVariables'])
    with temporary_file('w') as f:
        _setup_shared_config(
            f, case_configuration['sharedConfigFile'], environ
        )
        with mock.patch('os.environ', environ):
            session = botocore.session.Session()
            if 'maxRetries' in case_configuration:
                _setup_max_retry_attempts(session, case_configuration)
            yield session


def _setup_shared_config(fileobj, shared_config_options, environ):
    fileobj.write('[default]\n')
    for key, value in shared_config_options.items():
        fileobj.write(f'{key} = {value}\n')
    fileobj.flush()
    environ['AWS_CONFIG_FILE'] = fileobj.name


def _setup_max_retry_attempts(session, case_configuration):
    config = botocore.config.Config(
        retries={'max_attempts': case_configuration['maxRetries']}
    )
    session.set_default_client_config(config)


def _run_test_case(case):
    with MonitoringListener() as listener:
        with _configured_session(
            case['configuration'], listener.port
        ) as session:
            for api_call in case['apiCalls']:
                _make_api_call(session, api_call)
    assert listener.received_events == case['expectedMonitoringEvents']


def _make_api_call(session, api_call):
    client = session.create_client(
        api_call['serviceId'].lower().replace(' ', '')
    )
    operation_name = api_call['operationName']
    client_method = getattr(client, xform_name(operation_name))
    with _stubbed_http_layer(client, api_call['attemptResponses']):
        try:
            client_method(**api_call['params'])
        except EXPECTED_EXCEPTIONS_THROWN:
            pass


@contextlib.contextmanager
def _stubbed_http_layer(client, attempt_responses):
    with ClientHTTPStubber(client) as stubber:
        _add_stubbed_responses(stubber, attempt_responses)
        yield


def _add_stubbed_responses(stubber, attempt_responses):
    for attempt_response in attempt_responses:
        if 'sdkException' in attempt_response:
            sdk_exception = attempt_response['sdkException']
            _add_sdk_exception(
                stubber, sdk_exception['message'], sdk_exception['isRetryable']
            )
        else:
            _add_stubbed_response(stubber, attempt_response)


def _add_sdk_exception(stubber, message, is_retryable):
    if is_retryable:
        stubber.responses.append(RetryableException(message=message))
    else:
        stubber.responses.append(NonRetryableException(message))


def _add_stubbed_response(stubber, attempt_response):
    headers = attempt_response['responseHeaders']
    status_code = attempt_response['httpStatus']
    if 'errorCode' in attempt_response:
        error = {
            '__type': attempt_response['errorCode'],
            'message': attempt_response['errorMessage'],
        }
        content = json.dumps(error).encode('utf-8')
    else:
        content = b'{}'
    stubber.add_response(status=status_code, headers=headers, body=content)


class MonitoringListener(threading.Thread):
    _PACKET_SIZE = 1024 * 8

    def __init__(self, port=0):
        threading.Thread.__init__(self)
        self._socket = None
        self.port = port
        self.received_events = []

    def __enter__(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.bind(('127.0.0.1', self.port))
        # The socket may have been assigned to an unused port so we
        # reset the port member after binding.
        self.port = self._socket.getsockname()[1]
        self.start()
        return self

    def __exit__(self, *args):
        self._socket.sendto(b'', ('127.0.0.1', self.port))
        self.join()
        self._socket.close()

    def run(self):
        logger.debug('Started listener')
        while True:
            data = self._socket.recv(self._PACKET_SIZE)
            logger.debug('Received: %s', data.decode('utf-8'))
            if not data:
                return
            self.received_events.append(json.loads(data.decode('utf-8')))



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from tests.functional.docs import BaseDocsFunctionalTest


class TestCloudFormationDocs(BaseDocsFunctionalTest):
    def test_get_template_response_documented_as_dict(self):
        content = self.get_docstring_for_method(
            'cloudformation', 'get_template'
        )
        # String return type should be gone
        self.assert_not_contains_line("(*string*) --", content)
        # Check for template body returning a dict
        self.assert_contains_line("(*dict*) --", content)
        # Check the specifics of the returned dict
        self.assert_contains_line('{}', content)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

import botocore.session
from botocore.exceptions import WaiterError
from tests import random_chars, unittest


@pytest.mark.slow
class TestWaiterForDynamoDB(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client('dynamodb', 'us-west-2')

    def test_create_table_and_wait(self):
        table_name = 'botocoretest-%s' % random_chars(10)
        self.client.create_table(
            TableName=table_name,
            ProvisionedThroughput={
                "ReadCapacityUnits": 5,
                "WriteCapacityUnits": 5,
            },
            KeySchema=[{"AttributeName": "foo", "KeyType": "HASH"}],
            AttributeDefinitions=[
                {"AttributeName": "foo", "AttributeType": "S"}
            ],
        )
        self.addCleanup(self.client.delete_table, TableName=table_name)
        waiter = self.client.get_waiter('table_exists')
        waiter.wait(TableName=table_name)
        parsed = self.client.describe_table(TableName=table_name)
        self.assertEqual(parsed['Table']['TableStatus'], 'ACTIVE')


class TestCanGetWaitersThroughClientInterface(unittest.TestCase):
    def test_get_ses_waiter(self):
        # We're checking this because ses is not the endpoint prefix
        # for the service, it's email.  We want to make sure this does
        # not affect the lookup process.
        session = botocore.session.get_session()
        client = session.create_client('ses', 'us-east-1')
        # If we have at least one waiter in the list, we know that we have
        # actually loaded the waiters and this test has passed.
        self.assertTrue(len(client.waiter_names) > 0)


class TestMatchersWithErrors(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client(
            'ec2', region_name='us-west-2'
        )

    def test_dont_search_on_error_responses(self):
        """Test that InstanceExists can handle a nonexistent instance."""
        waiter = self.client.get_waiter('instance_exists')
        waiter.config.max_attempts = 1
        with self.assertRaises(WaiterError):
            waiter.wait(InstanceIds=['i-12345'])
